//Code for the Arduino/Esp
//Version: 10
//Contains: i2c, deepsleep, wifi, OTA and delay (with auth), mqtt (with auth), led control

#include <Wire.h>
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <PubSubClient.h>

// WIFI variables and settings
#ifndef STASSID
#define STASSID "**********"
#define STAPSK  "**********"
#endif

const char* ssid = STASSID;
const char* password = STAPSK;

// Setting a Static IP address
IPAddress local_IP(192, 168, 0, 138);
IPAddress gateway(192, 168, 0, 1);
IPAddress subnet(255, 255, 255, 0);

// External LED settings when auth fails
const int LED = 14;

// I2C array & function
int arr_a[] = {0, 0, 0, 0}; //Hardcoded password array, change this as neeed, the attiny must match this code

int arr_b[4];  //an array to hold all 4 bytes from attiny i2c slave
byte arrayIndex = 0;

// array checking
boolean array_cmp(int *a, int *b, int len_a, int len_b){
      int n;

      // test each element to be the same. if not, return false
      for (n=0;n<len_a;n++) if (a[n]!=b[n]) return false;

      //ok, if we have not returned yet, they are equal
      return true;
}

// MQTT settings
const char* mqtt_server = "192.168.0.130";
const char* key_topic = "******";
const char* clientID = "ESP8266Client";
const char* usrName = "******";
const char* passWD = "******";

// Initialise the WiFi and MQTT Client objects
WiFiClient wifiClient;
// 1883 is the listener port for the Broker
PubSubClient client(mqtt_server, 1883, wifiClient);

void setup() {
  pinMode (LED, OUTPUT);
  pinMode(LED_BUILTIN, OUTPUT); // initialize digital pin LED_BUILTIN as an output.
  digitalWrite(LED_BUILTIN, HIGH);
  Wire.begin(); // begin i2c bus
  Serial.begin(115200);
  Serial.println("\nBooting");
  Serial.println("Code Version: 10");
  WiFi.mode(WIFI_STA);
  // Configures static IP address
  if (!WiFi.config(local_IP, gateway, subnet)) {
    Serial.println("STA Failed to configure");
  }
  WiFi.begin(ssid, password);
  while (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Serial.println("Connection Failed! Rebooting...");
    delay(5000);
    ESP.restart();
  }
  Serial.println("Wifi done");
  // Connect to MQTT Broker
  // client.connect returns a boolean value to let us know if the connection was successful.
  if (client.connect(clientID, usrName, passWD)) {
    Serial.println("Connected to MQTT Broker!");
  }
  else {
    Serial.println("Connection to MQTT Broker failed...");
  }

  // Port defaults to 8266
  // ArduinoOTA.setPort(8266);

  // Hostname defaults to esp8266-[ChipID]
  // ArduinoOTA.setHostname("myesp8266");

  // No authentication by default
  ArduinoOTA.setPassword("********");

  // Password can be set with it's md5 value as well
  // MD5(admin) = 21232f297a57a5a743894a0e4a801fc3
  // ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3");
  Serial.println("OTA setup begin");
  ArduinoOTA.onStart([]() {
    String type;
    if (ArduinoOTA.getCommand() == U_FLASH) {
      type = "sketch";
    } else { // U_FS
      type = "filesystem";
    }

    // NOTE: if updating FS this would be the place to unmount FS using FS.end()
    Serial.println("Start updating " + type);
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) {
      Serial.println("Auth Failed");
    } else if (error == OTA_BEGIN_ERROR) {
      Serial.println("Begin Failed");
    } else if (error == OTA_CONNECT_ERROR) {
      Serial.println("Connect Failed");
    } else if (error == OTA_RECEIVE_ERROR) {
      Serial.println("Receive Failed");
    } else if (error == OTA_END_ERROR) {
      Serial.println("End Failed");
    }
  });
  ArduinoOTA.begin();
  Serial.println("Ready");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
  Serial.println("OTA setup DONE");
}

void loop()
{
 ArduinoOTA.handle();
 Serial.println("Starting I2C process");
 Wire.requestFrom(4, 4); // request 4 byte from slave device address 4 (addres, byte)
 
while(Wire.available()) // slave may send less than requested
 {
 int i = Wire.read(); // receive a byte as character
 Serial.println(i); // print the character
 arr_b[arrayIndex] = i; // add the byte to the array
 arrayIndex++; // increase array index
 }
 // These two for loops are used for debugging to check both array values
 Serial.println("---------a-----------\n");
 for (int y = 0; y < 4; y++)
  {
     Serial.println(arr_a[y]);
  }
  Serial.println("---------b-----------\n");
 for (int y = 0; y < 4; y++)
  {
     Serial.println(arr_b[y]);
  }
  Serial.println("---------END-----------\n");
  // This if/else statement is used to check if the arrays are the same the run an action
  if (array_cmp(arr_a, arr_b, 4, 4) == true){
            // do this if they are equal
            Serial.println("Access Granted: array matches");
            //digitalWrite(LED_BUILTIN, LOW); // Light up LED for visual indecation
            // Send to MQTT broker
            if (client.publish(key_topic, "**********")) {
              Serial.println("Key sent! - Granted");
              }
            client.disconnect();
            memset(arr_b, 0, sizeof(arr_b)); // Clear the array that came from the attiny
            arrayIndex = 0; // Restart the index counter
            //delay(1000);
            digitalWrite(LED_BUILTIN, LOW);
            delay(500);
            digitalWrite(LED_BUILTIN, HIGH);
            delay(500);
            digitalWrite(LED_BUILTIN, LOW);
            delay(500);
            digitalWrite(LED_BUILTIN, HIGH);
            delay(500);
            digitalWrite(LED_BUILTIN, LOW);
            delay(500);
            digitalWrite(LED_BUILTIN, HIGH);
            
      }else{
            // do this if they are different
            Serial.println("Access Denied: array missmatch");
            // Send to MQTT broker
            if (client.publish(key_topic, "*********")) {
              Serial.println("Key sent! - Denied");
              }
            client.disconnect();
            memset(arr_b, 0, sizeof(arr_b)); // Clear the array that came from the attiny
            arrayIndex = 0; // Restart the index counter
            digitalWrite(LED, HIGH);
            delay(500);
            digitalWrite(LED, LOW);
            delay(500);
            digitalWrite(LED, HIGH);
            delay(500);
            digitalWrite(LED, LOW);
            delay(500);
            digitalWrite(LED, HIGH);
            delay(500);
            digitalWrite(LED, LOW);
            
      }
 
Serial.println("Waiting 15 seconds for possible OTA");
// A method of waiting 15 seconds without using delay as delay will block incoming OTA request
uint32_t moment = millis();
while (millis() - moment < 15000) {
  ArduinoOTA.handle();
  yield();
}
Serial.println("sleeping till externally woken");
ESP.deepSleep(0);  // Put the ESP to sleep indefinatly till externally woken by a reset
}
